package magic.ui.screen.images.download;

import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.Proxy;
import java.net.URL;
import java.nio.file.Files;
import java.time.LocalDateTime;
import java.util.List;
import java.util.concurrent.CancellationException;
import java.util.concurrent.ExecutionException;
import javax.swing.SwingUtilities;
import javax.swing.SwingWorker;
import magic.data.CardImageFile;
import magic.data.DownloadableFile;
import magic.data.GeneralConfig;
import magic.data.ImagesDownloadList;
import magic.model.MagicCardDefinition;
import magic.ui.CardTextLanguage;
import magic.ui.MagicImages;
import magic.ui.MagicLogFile;
import magic.ui.helpers.UrlHelper;
import magic.utility.MagicFileSystem;
import magic.utility.MagicSystem;

class DownloadWorker extends SwingWorker<Void, Integer> {

    private static final MagicLogFile missingLog = new MagicLogFile("downloaded-images");

    private final Proxy proxy;
    private ImagesDownloadList downloadList;
    private final IDownloadListener listener;
    private final CardTextLanguage textLanguage;
    private final CardImageDisplayMode displayMode;
    private boolean updateDownloadDate = true;
    private boolean isLogging = true;
    private int serverBusyCooldown = 1; // in millisecs

    DownloadWorker(IDownloadListener aListener, CardTextLanguage aLanguage, CardImageDisplayMode aMode) {
        this.listener = aListener;
        this.textLanguage = aLanguage;
        this.displayMode = aMode;
        this.proxy = GeneralConfig.getInstance().getProxy();
    }

    @Override
    protected Void doInBackground() throws MalformedURLException {
        this.downloadList = ScanWorker.getImagesDownloadList((IScanListener)listener, displayMode);
        doDownloadImages(textLanguage);
        return null;
    }

    @Override
    protected void done() {
        try {
            get();
        } catch (ExecutionException | InterruptedException ex) {
            throw new RuntimeException(ex);
        } catch (CancellationException ex) {
            System.err.println("ImageDownloadWorker cancelled.");
        }
        listener.setButtonState(false);
        resetProgressBar();

        MagicImages.clearCache();
        listener.buildDownloadImagesList();
    }

    @Override
    protected void process(List<Integer> chunks) {
        final int countInteger = chunks.get(chunks.size() - 1);
        if (!isCancelled()) {
            listener.showProgress(countInteger, downloadList.size());
        }
    }

    private boolean isLoggingOn() {
        return isLogging && MagicSystem.isDevMode();
    }

    /**
     * Experimental, currently only available in devMode and missing mode.
     */
    private void doLog(String cardName, CardTextLanguage aLang, URL cardUrl) {
        if (isLoggingOn()) {
            synchronized (missingLog) {
                try {
                    missingLog.log(cardName, aLang, cardUrl);
                } catch (IOException ex) {
                    System.err.println("doLog: " + ex);
                    isLogging = false;
                }
            }
        }
    }

    /**
     * Initializes log files for capturing useful info (DEVMODE ONLY!).
     */
    private void initializeLogFiles() {
        if (isLoggingOn()) {
            synchronized (missingLog) {
                try {
                    missingLog.deleteLogFileIfExists();
                    missingLog.log("# " + LocalDateTime.now());
                    missingLog.log("#");
                } catch (IOException ex) {
                    System.err.println("initializeLogFiles: " + ex);
                    isLogging = false;
                }
            }
        }
    }

    /**
     * Gets alternate magiccards.info image url for given card text language.
     * If it fails then displays reason in error log panel and returns null url.
     */
    private URL getAlternateUrl(DownloadableFile aFile, CardTextLanguage textLang) {
        try {
            return UrlHelper.getAlternateMagicCardsImageUrl(aFile.getUrl(), textLang);
        } catch (MalformedURLException ex) {
            listener.setMessage(String.format("%s [%s]", ex.toString(), aFile.getUrl()));
            return null;
        }
    }

    private void setServerBusyCooldown(String errmsg) {
        if (errmsg.contains("HTTP response code: 503")) {
            serverBusyCooldown += 100;
        }
    }

    /**
     * Attempts to download alternate card image.
     * If it fails displays reason in error log panel.
     */
    private boolean tryAlternateDownload(final DownloadableFile aFile, CardTextLanguage aLang) {
        try {
            aFile.doDownload(proxy);
        } catch (IOException ex) {
            listener.setMessage(String.format("%s [%s (%s)]", ex.toString(), aFile.getFilename(), aLang));
            setServerBusyCooldown(ex.toString());
            return false;
        }
        return true;
    }

    private boolean downloadAlternateCardImage(final CardImageFile aFile, CardTextLanguage aLang) {
        final URL altUrl = getAlternateUrl(aFile, aLang);
        if (UrlHelper.isUrlValid(altUrl)) {
            if (tryAlternateDownload(new DownloadableFile(aFile.getLocalFile(), altUrl), aLang)) {
                doLog(aFile.getCardName(), aLang, altUrl);
                return true;
            }
        }
        return false;
    }

    private boolean doDeleteLocalImageFile(File aFile) {
        try {
            Files.deleteIfExists(aFile.toPath());
        } catch (IOException ex) {
            listener.setMessage(String.format("%s [%s]", ex.toString(), aFile.getName()));
            return false;
        }
        return true;
    }

    /**
     * Attempts to download default card image.
     * If it fails displays reason in error log panel.
     */
    private boolean tryDefaultDownload(final DownloadableFile aFile) {
        try {
            aFile.doDownload(proxy);
        } catch (IOException ex) {
            // if local file exists then download was triggered by the
            // 'image_updated' script property. But download failed so
            // remove local image so it becomes 'missing' instead.
            updateDownloadDate = displayMode != CardImageDisplayMode.PROXY && doDeleteLocalImageFile(aFile.getLocalFile());
            listener.setMessage(String.format("%s [%s]", ex.toString(), aFile.getFilename()));
            setServerBusyCooldown(ex.toString());
            return false;
        }
        return true;
    }

    private void downloadDefaultCardImage(final CardImageFile aFile) {
        if (tryDefaultDownload(aFile)) {
            doLog(aFile.getCardName(), CardTextLanguage.ENGLISH, null);
        }
    }

    private boolean doDownloadCroppedImage(CardImageFile imageFile) throws MalformedURLException {
        final MagicCardDefinition card = imageFile.getCard();
        final File local = MagicFileSystem.getCroppedCardImage(imageFile.getCard());
        if (card.getImageURL().contains("magiccards.info/scans/")) {
            final URL remote = new URL(card.getImageURL().replace("/scans/", "/crop/"));
            if (UrlHelper.isUrlValid(remote)) {
                if (tryDefaultDownload(new DownloadableFile(local, remote))) {
                    doLog(card.getCardTextName(), CardTextLanguage.ENGLISH, remote);
                    return true;
                }
            }
        }
        return false;
    }

    private void doDownloadPrintedImage(CardImageFile imageFile, CardTextLanguage textLang) throws MalformedURLException {
        if (textLang.isEnglish() || !downloadAlternateCardImage(imageFile, textLang)) {
            downloadDefaultCardImage(imageFile);
        }
    }

    private void doPause(int millisecs) {
        try {
            Thread.sleep(millisecs);
        } catch (InterruptedException ex) {
            System.err.println(ex);
        }
    }

    private void doDownloadImages(CardTextLanguage textLang) throws MalformedURLException {

        assert !SwingUtilities.isEventDispatchThread();

        initializeLogFiles();
        int fileCount = 0;
        updateDownloadDate = displayMode != CardImageDisplayMode.PROXY;

        for (DownloadableFile dFile : downloadList) {

            final CardImageFile imageFile = (CardImageFile) dFile;

            if (displayMode == CardImageDisplayMode.PROXY) {
                if (!doDownloadCroppedImage(imageFile)) {
                    doDownloadPrintedImage(imageFile, textLang);
                }
            } else {
                doDownloadPrintedImage(imageFile, textLang);
            }
            doPause(serverBusyCooldown);

            fileCount++;

            if (isCancelled()) {
                return;
            } else {
                publish(new Integer(fileCount));
            }
        }

        if (updateDownloadDate) {
            listener.doCustomActionAfterDownload();
        }

    }

    private void resetProgressBar() {
        assert SwingUtilities.isEventDispatchThread();
        listener.resetProgress();
    }

}
